pub struct MemoryAccessOptions(/* private fields */);Expand description
Options for controlling memory access monitoring.
These options can be used to fine-tune the behavior of memory access monitoring, allowing you to ignore certain types of memory accesses and improve performance.
Implementations
Sourceimpl MemoryAccessOptions
impl MemoryAccessOptions
Sourcepub const IGNORE_PAGE_WALK_UPDATES: MemoryAccessOptions
Available on crate features utils and injector only.
pub const IGNORE_PAGE_WALK_UPDATES: MemoryAccessOptions
utils and injector only.Ignore page-walk updates caused by the CPU.
When this flag is set, memory accesses that are solely the result of
CPU-initiated page-table walks will not trigger an EventMemoryAccess.
This is useful for filtering out irrelevant events when monitoring
page-table modifications.
Notes
This option is only effective when the [MemoryAccess:W] is not set.
Sourceimpl MemoryAccessOptions
impl MemoryAccessOptions
Sourcepub const fn empty() -> MemoryAccessOptions
Available on crate features utils and injector only.
pub const fn empty() -> MemoryAccessOptions
utils and injector only.Get a flags value with all bits unset.
Sourcepub const fn all() -> MemoryAccessOptions
Available on crate features utils and injector only.
pub const fn all() -> MemoryAccessOptions
utils and injector only.Get a flags value with all known bits set.
Sourcepub const fn bits(&self) -> u8
Available on crate features utils and injector only.
pub const fn bits(&self) -> u8
utils and injector only.Get the underlying bits value.
The returned value is exactly the bits set in this flags value.
Sourcepub const fn from_bits(bits: u8) -> Option<MemoryAccessOptions>
Available on crate features utils and injector only.
pub const fn from_bits(bits: u8) -> Option<MemoryAccessOptions>
utils and injector only.Convert from a bits value.
This method will return None if any unknown bits are set.
Sourcepub const fn from_bits_truncate(bits: u8) -> MemoryAccessOptions
Available on crate features utils and injector only.
pub const fn from_bits_truncate(bits: u8) -> MemoryAccessOptions
utils and injector only.Convert from a bits value, unsetting any unknown bits.
Sourcepub const fn from_bits_retain(bits: u8) -> MemoryAccessOptions
Available on crate features utils and injector only.
pub const fn from_bits_retain(bits: u8) -> MemoryAccessOptions
utils and injector only.Convert from a bits value exactly.
Sourcepub fn from_name(name: &str) -> Option<MemoryAccessOptions>
Available on crate features utils and injector only.
pub fn from_name(name: &str) -> Option<MemoryAccessOptions>
utils and injector only.Get a flags value with the bits of a flag with the given name set.
This method will return None if name is empty or doesn’t
correspond to any named flag.
Sourcepub const fn is_empty(&self) -> bool
Available on crate features utils and injector only.
pub const fn is_empty(&self) -> bool
utils and injector only.Whether all bits in self are unset.
Sourcepub const fn is_all(&self) -> bool
Available on crate features utils and injector only.
pub const fn is_all(&self) -> bool
utils and injector only.Whether all known bits in this flags value are set.
Sourcepub const fn intersects(&self, other: MemoryAccessOptions) -> bool
Available on crate features utils and injector only.
pub const fn intersects(&self, other: MemoryAccessOptions) -> bool
utils and injector only.Whether any set bits in other are also set in self.
Sourcepub const fn contains(&self, other: MemoryAccessOptions) -> bool
Available on crate features utils and injector only.
pub const fn contains(&self, other: MemoryAccessOptions) -> bool
utils and injector only.Whether all set bits in other are also set in self.
Sourcepub fn insert(&mut self, other: MemoryAccessOptions)
Available on crate features utils and injector only.
pub fn insert(&mut self, other: MemoryAccessOptions)
utils and injector only.The bitwise or (|) of the bits in self and other.
Sourcepub fn remove(&mut self, other: MemoryAccessOptions)
Available on crate features utils and injector only.
pub fn remove(&mut self, other: MemoryAccessOptions)
utils and injector only.The intersection of self with the complement of other (&!).
This method is not equivalent to self & !other when other has unknown bits set.
remove won’t truncate other, but the ! operator will.
Sourcepub fn toggle(&mut self, other: MemoryAccessOptions)
Available on crate features utils and injector only.
pub fn toggle(&mut self, other: MemoryAccessOptions)
utils and injector only.The bitwise exclusive-or (^) of the bits in self and other.
Sourcepub fn set(&mut self, other: MemoryAccessOptions, value: bool)
Available on crate features utils and injector only.
pub fn set(&mut self, other: MemoryAccessOptions, value: bool)
utils and injector only.Call insert when value is true or remove when value is false.
Sourcepub const fn intersection(
self,
other: MemoryAccessOptions,
) -> MemoryAccessOptions
Available on crate features utils and injector only.
pub const fn intersection( self, other: MemoryAccessOptions, ) -> MemoryAccessOptions
utils and injector only.The bitwise and (&) of the bits in self and other.
Sourcepub const fn union(self, other: MemoryAccessOptions) -> MemoryAccessOptions
Available on crate features utils and injector only.
pub const fn union(self, other: MemoryAccessOptions) -> MemoryAccessOptions
utils and injector only.The bitwise or (|) of the bits in self and other.
Sourcepub const fn difference(self, other: MemoryAccessOptions) -> MemoryAccessOptions
Available on crate features utils and injector only.
pub const fn difference(self, other: MemoryAccessOptions) -> MemoryAccessOptions
utils and injector only.The intersection of self with the complement of other (&!).
This method is not equivalent to self & !other when other has unknown bits set.
difference won’t truncate other, but the ! operator will.
Sourcepub const fn symmetric_difference(
self,
other: MemoryAccessOptions,
) -> MemoryAccessOptions
Available on crate features utils and injector only.
pub const fn symmetric_difference( self, other: MemoryAccessOptions, ) -> MemoryAccessOptions
utils and injector only.The bitwise exclusive-or (^) of the bits in self and other.
Sourcepub const fn complement(self) -> MemoryAccessOptions
Available on crate features utils and injector only.
pub const fn complement(self) -> MemoryAccessOptions
utils and injector only.The bitwise negation (!) of the bits in self, truncating the result.
Sourceimpl MemoryAccessOptions
impl MemoryAccessOptions
Sourcepub const fn iter(&self) -> Iter<MemoryAccessOptions> ⓘ
Available on crate features utils and injector only.
pub const fn iter(&self) -> Iter<MemoryAccessOptions> ⓘ
utils and injector only.Yield a set of contained flags values.
Each yielded flags value will correspond to a defined named flag. Any unknown bits will be yielded together as a final flags value.
Sourcepub const fn iter_names(&self) -> IterNames<MemoryAccessOptions> ⓘ
Available on crate features utils and injector only.
pub const fn iter_names(&self) -> IterNames<MemoryAccessOptions> ⓘ
utils and injector only.Yield a set of contained named flags values.
This method is like iter, except only yields bits in contained named flags.
Any unknown bits, or bits not corresponding to a contained flag will not be yielded.