VmiDriver

Trait VmiDriver 

pub trait VmiDriver: 'static {
    type Architecture: Architecture;

    // Required method
    fn info(&self) -> Result<VmiInfo, VmiError>;
}
Base trait for all VMI driver sub-traits.

This trait provides the associated Architecture type and the fundamental info() method for querying VM metadata.

The 'static lifetime is required in order to use the driver with the VmiOs enumerators.

Required Associated Types

type Architecture: Architecture

The architecture supported by the driver.

Required Methods

fn info(&self) -> Result<VmiInfo, VmiError>

Returns information about the virtual machine.

Dyn Compatibility

This trait is dyn compatible.

In older versions of Rust, dyn compatibility was called "object safety".

Implementors

Source

impl<Arch> VmiDriver for VmiKdmpDriver<Arch>
where Arch: ArchAdapter,

Source

impl<Arch> VmiDriver for VmiXenCoreDumpDriver<Arch>
where Arch: ArchAdapter,

Source

impl<Arch> VmiDriver for VmiXenDriver<Arch>
where Arch: ArchAdapter,