Module utils

Module utils 

Available on crate feature utils only.
VMI utilities.

Modules

bpmbpm
Breakpoint management.
bridgebridge
Guest-host communication bridge for VMI.
injectorinjector
Code injection functionality.
interceptorinterceptor
Simple software breakpoint management.
ptmptm
Page table monitoring.
reactorreactor
Symbol-driven VMI event loop.
resolverresolver
Locates loaded modules in memory and extracts the information needed to fetch their debug symbols.

Enums

Representation
Representation of memory for hexdump.

Functions

hexdump
Print a hexdump of memory at the given address.