pub struct CallBuilder { /* private fields */ }Available on crate features
injector and utils only.Expand description
A builder struct for constructing calls to system functions, encapsulating the function’s address and its required arguments.
Implementations
Sourceimpl CallBuilder
impl CallBuilder
Sourcepub fn new(function_address: Va) -> CallBuilder
pub fn new(function_address: Va) -> CallBuilder
Creates a new CallBuilder instance for a specified function address.
Sourcepub fn with_argument(self, argument: impl Into<Argument>) -> CallBuilder
pub fn with_argument(self, argument: impl Into<Argument>) -> CallBuilder
Adds an argument to the function call being built.