pub struct InjectorHandler<Os, Mode, T, Bridge = ()>where
Os: InjectorExecutionAdapter<Mode, T, Bridge>,
Mode: ExecutionMode,
Bridge: BridgeDispatch<Os, u64>,{ /* private fields */ }Available on crate features
injector and utils only.Expand description
Generic injector handler that delegates to an OS- and mode-specific implementation.
Prefer the KernelInjectorHandler and UserInjectorHandler type
aliases for the common case without a bridge. When a custom
BridgeDispatch is needed, use this type directly with an explicit
Bridge parameter.
Implementations
Sourceimpl<Os, Mode, T, Bridge> InjectorHandler<Os, Mode, T, Bridge>where
Os: InjectorExecutionAdapter<Mode, T, Bridge>,
Mode: ExecutionMode,
Bridge: BridgeDispatch<Os, u64>,
impl<Os, Mode, T, Bridge> InjectorHandler<Os, Mode, T, Bridge>where
Os: InjectorExecutionAdapter<Mode, T, Bridge>,
Mode: ExecutionMode,
Bridge: BridgeDispatch<Os, u64>,
Sourcepub fn new(
vmi: &VmiSession<'_, Os>,
recipe: Recipe<Os, T>,
) -> Result<InjectorHandler<Os, Mode, T, Bridge>, VmiError>where
Bridge: Default,
pub fn new(
vmi: &VmiSession<'_, Os>,
recipe: Recipe<Os, T>,
) -> Result<InjectorHandler<Os, Mode, T, Bridge>, VmiError>where
Bridge: Default,
Creates a new injector handler with a default (no-op) bridge.
Examples found in repository?
examples/windows-recipe-messagebox.rs (lines 94-100)
65fn main() -> Result<(), Error> {
66 let session = common::create_vmi_session()?;
67
68 let explorer_pid = {
69 // This block is used to drop the pause guard after the PID is found.
70 // If the `session.handle()` would be called with the VM paused, no
71 // events would be triggered.
72 let paused = session.pause_guard()?;
73
74 let vmi = paused.state();
75
76 let explorer = match vmi.os().find_process("explorer.exe")? {
77 Some(explorer) => explorer,
78 None => {
79 tracing::error!("explorer.exe not found");
80 return Ok(());
81 }
82 };
83
84 tracing::info!(
85 pid = %explorer.id()?,
86 object = %explorer.object()?,
87 "found explorer.exe"
88 );
89
90 explorer.id()?
91 };
92
93 session.handle(|session| {
94 UserInjectorHandler::new(
95 session,
96 recipe_factory(MessageBox::new(
97 "Hello, World!",
98 "This is a message box from the VMI!",
99 )),
100 )?
101 .with_pid(explorer_pid)
102 })?;
103
104 Ok(())
105}More examples
examples/windows-recipe-writefile.rs (lines 234-240)
205fn main() -> Result<(), Error> {
206 let session = common::create_vmi_session()?;
207
208 let explorer_pid = {
209 // This block is used to drop the pause guard after the PID is found.
210 // If the `session.handle()` would be called with the VM paused, no
211 // events would be triggered.
212 let paused = session.pause_guard()?;
213
214 let vmi = paused.state();
215
216 let explorer = match vmi.os().find_process("explorer.exe")? {
217 Some(explorer) => explorer,
218 None => {
219 tracing::error!("explorer.exe not found");
220 return Ok(());
221 }
222 };
223
224 tracing::info!(
225 pid = %explorer.id()?,
226 object = %explorer.object()?,
227 "found explorer.exe"
228 );
229
230 explorer.id()?
231 };
232
233 session.handle(|session| {
234 UserInjectorHandler::new(
235 session,
236 recipe_factory(GuestFile::new(
237 "C:\\Users\\John\\Desktop\\test.txt",
238 "Hello, World!".as_bytes(),
239 )),
240 )?
241 .with_pid(explorer_pid)
242 })?;
243
244 Ok(())
245}examples/windows-recipe-writefile-advanced.rs (lines 337-343)
303fn main() -> Result<(), Error> {
304 let session = common::create_vmi_session()?;
305
306 let explorer_pid = {
307 // This block is used to drop the pause guard after the PID is found.
308 // If the `session.handle()` would be called with the VM paused, no
309 // events would be triggered.
310 let paused = session.pause_guard()?;
311
312 let vmi = paused.state();
313
314 let explorer = match vmi.os().find_process("explorer.exe")? {
315 Some(explorer) => explorer,
316 None => {
317 tracing::error!("explorer.exe not found");
318 return Ok(());
319 }
320 };
321
322 tracing::info!(
323 pid = %explorer.id()?,
324 object = %explorer.object()?,
325 "found explorer.exe"
326 );
327
328 explorer.id()?
329 };
330
331 let mut content = Vec::new();
332 for c in 'A'..='Z' {
333 content.extend((0..2049).map(|_| c as u8).collect::<Vec<_>>());
334 }
335
336 session.handle(|session| {
337 UserInjectorHandler::new(
338 session,
339 recipe_factory(GuestFile::new(
340 "C:\\Users\\John\\Desktop\\test.txt",
341 content,
342 )),
343 )?
344 .with_pid(explorer_pid)
345 })?;
346
347 Ok(())
348}Sourcepub fn with_bridge(
vmi: &VmiSession<'_, Os>,
bridge: Bridge,
recipe: Recipe<Os, T>,
) -> Result<InjectorHandler<Os, Mode, T, Bridge>, VmiError>
pub fn with_bridge( vmi: &VmiSession<'_, Os>, bridge: Bridge, recipe: Recipe<Os, T>, ) -> Result<InjectorHandler<Os, Mode, T, Bridge>, VmiError>
Creates a new injector handler with a custom bridge for guest-host communication.
Sourcepub fn with_pid(
self,
pid: ProcessId,
) -> Result<InjectorHandler<Os, Mode, T, Bridge>, VmiError>
pub fn with_pid( self, pid: ProcessId, ) -> Result<InjectorHandler<Os, Mode, T, Bridge>, VmiError>
Restricts injection to a specific process.
Examples found in repository?
examples/windows-recipe-messagebox.rs (line 101)
65fn main() -> Result<(), Error> {
66 let session = common::create_vmi_session()?;
67
68 let explorer_pid = {
69 // This block is used to drop the pause guard after the PID is found.
70 // If the `session.handle()` would be called with the VM paused, no
71 // events would be triggered.
72 let paused = session.pause_guard()?;
73
74 let vmi = paused.state();
75
76 let explorer = match vmi.os().find_process("explorer.exe")? {
77 Some(explorer) => explorer,
78 None => {
79 tracing::error!("explorer.exe not found");
80 return Ok(());
81 }
82 };
83
84 tracing::info!(
85 pid = %explorer.id()?,
86 object = %explorer.object()?,
87 "found explorer.exe"
88 );
89
90 explorer.id()?
91 };
92
93 session.handle(|session| {
94 UserInjectorHandler::new(
95 session,
96 recipe_factory(MessageBox::new(
97 "Hello, World!",
98 "This is a message box from the VMI!",
99 )),
100 )?
101 .with_pid(explorer_pid)
102 })?;
103
104 Ok(())
105}More examples
examples/windows-recipe-writefile.rs (line 241)
205fn main() -> Result<(), Error> {
206 let session = common::create_vmi_session()?;
207
208 let explorer_pid = {
209 // This block is used to drop the pause guard after the PID is found.
210 // If the `session.handle()` would be called with the VM paused, no
211 // events would be triggered.
212 let paused = session.pause_guard()?;
213
214 let vmi = paused.state();
215
216 let explorer = match vmi.os().find_process("explorer.exe")? {
217 Some(explorer) => explorer,
218 None => {
219 tracing::error!("explorer.exe not found");
220 return Ok(());
221 }
222 };
223
224 tracing::info!(
225 pid = %explorer.id()?,
226 object = %explorer.object()?,
227 "found explorer.exe"
228 );
229
230 explorer.id()?
231 };
232
233 session.handle(|session| {
234 UserInjectorHandler::new(
235 session,
236 recipe_factory(GuestFile::new(
237 "C:\\Users\\John\\Desktop\\test.txt",
238 "Hello, World!".as_bytes(),
239 )),
240 )?
241 .with_pid(explorer_pid)
242 })?;
243
244 Ok(())
245}examples/windows-recipe-writefile-advanced.rs (line 344)
303fn main() -> Result<(), Error> {
304 let session = common::create_vmi_session()?;
305
306 let explorer_pid = {
307 // This block is used to drop the pause guard after the PID is found.
308 // If the `session.handle()` would be called with the VM paused, no
309 // events would be triggered.
310 let paused = session.pause_guard()?;
311
312 let vmi = paused.state();
313
314 let explorer = match vmi.os().find_process("explorer.exe")? {
315 Some(explorer) => explorer,
316 None => {
317 tracing::error!("explorer.exe not found");
318 return Ok(());
319 }
320 };
321
322 tracing::info!(
323 pid = %explorer.id()?,
324 object = %explorer.object()?,
325 "found explorer.exe"
326 );
327
328 explorer.id()?
329 };
330
331 let mut content = Vec::new();
332 for c in 'A'..='Z' {
333 content.extend((0..2049).map(|_| c as u8).collect::<Vec<_>>());
334 }
335
336 session.handle(|session| {
337 UserInjectorHandler::new(
338 session,
339 recipe_factory(GuestFile::new(
340 "C:\\Users\\John\\Desktop\\test.txt",
341 content,
342 )),
343 )?
344 .with_pid(explorer_pid)
345 })?;
346
347 Ok(())
348}Trait Implementations
Sourceimpl<Os, Mode, T, Bridge> VmiHandler<Os> for InjectorHandler<Os, Mode, T, Bridge>where
Os: InjectorExecutionAdapter<Mode, T, Bridge>,
Mode: ExecutionMode,
Bridge: BridgeDispatch<Os, u64>,
impl<Os, Mode, T, Bridge> VmiHandler<Os> for InjectorHandler<Os, Mode, T, Bridge>where
Os: InjectorExecutionAdapter<Mode, T, Bridge>,
Mode: ExecutionMode,
Bridge: BridgeDispatch<Os, u64>,
Sourcetype Output = <<Os as InjectorExecutionAdapter<Mode, T, Bridge>>::Handler as VmiHandler<Os>>::Output
type Output = <<Os as InjectorExecutionAdapter<Mode, T, Bridge>>::Handler as VmiHandler<Os>>::Output
The output type of the handler.
Sourcefn handle_event(
&mut self,
vmi: VmiContext<'_, Os>,
) -> VmiEventResponse<<Os as VmiOs>::Architecture>
fn handle_event( &mut self, vmi: VmiContext<'_, Os>, ) -> VmiEventResponse<<Os as VmiOs>::Architecture>
Called for each VMI event. Read more
Sourcefn poll(
&self,
) -> Option<<InjectorHandler<Os, Mode, T, Bridge> as VmiHandler<Os>>::Output>
fn poll( &self, ) -> Option<<InjectorHandler<Os, Mode, T, Bridge> as VmiHandler<Os>>::Output>
Checks if the handler has completed. Read more
Sourcefn handle_timeout(&mut self, _session: &VmiSession<'_, Os>)
fn handle_timeout(&mut self, _session: &VmiSession<'_, Os>)
Called when the event loop times out waiting for the next event. Read more
Sourcefn handle_interrupted(&mut self, _session: &VmiSession<'_, Os>)
fn handle_interrupted(&mut self, _session: &VmiSession<'_, Os>)
Called when the event loop is interrupted by a signal. Read more
Sourcefn cleanup(&mut self, _session: &VmiSession<'_, Os>)
fn cleanup(&mut self, _session: &VmiSession<'_, Os>)
Called once before the session tears down monitoring. Read more