pub struct InjectorHandler<Os, Mode, T, Bridge = ()>where
Os: InjectorExecutionAdapter<Mode, T, Bridge>,
Mode: ExecutionMode,
Bridge: BridgeDispatch<Os, u64>,{ /* private fields */ }Available on crate features
injector and utils only.Expand description
Generic injector handler that delegates to an OS- and mode-specific implementation.
Prefer the KernelInjectorHandler and UserInjectorHandler type
aliases for the common case without a bridge. When a custom
BridgeDispatch is needed, use this type directly with an explicit
Bridge parameter.
Implementations
Sourceimpl<Os, Mode, T, Bridge> InjectorHandler<Os, Mode, T, Bridge>where
Os: InjectorExecutionAdapter<Mode, T, Bridge>,
Mode: ExecutionMode,
Bridge: BridgeDispatch<Os, u64>,
impl<Os, Mode, T, Bridge> InjectorHandler<Os, Mode, T, Bridge>where
Os: InjectorExecutionAdapter<Mode, T, Bridge>,
Mode: ExecutionMode,
Bridge: BridgeDispatch<Os, u64>,
Sourcepub fn new(
vmi: &VmiSession<'_, Os>,
recipe: Recipe<Os, T>,
) -> Result<InjectorHandler<Os, Mode, T, Bridge>, VmiError>where
Bridge: Default,
pub fn new(
vmi: &VmiSession<'_, Os>,
recipe: Recipe<Os, T>,
) -> Result<InjectorHandler<Os, Mode, T, Bridge>, VmiError>where
Bridge: Default,
Creates a new injector handler with a default (no-op) bridge.
Examples found in repository?
examples/windows-recipe-messagebox.rs (lines 93-99)
64fn main() -> Result<(), Box<dyn std::error::Error>> {
65 let (session, _profile) = common::create_vmi_session()?;
66
67 let explorer_pid = {
68 // This block is used to drop the pause guard after the PID is found.
69 // If the `session.handle()` would be called with the VM paused, no
70 // events would be triggered.
71 let paused = session.pause_guard()?;
72
73 let vmi = paused.state();
74
75 let explorer = match common::find_process(&vmi, "explorer.exe")? {
76 Some(explorer) => explorer,
77 None => {
78 tracing::error!("explorer.exe not found");
79 return Ok(());
80 }
81 };
82
83 tracing::info!(
84 pid = %explorer.id()?,
85 object = %explorer.object()?,
86 "found explorer.exe"
87 );
88
89 explorer.id()?
90 };
91
92 session.handle(|session| {
93 UserInjectorHandler::new(
94 session,
95 recipe_factory(MessageBox::new(
96 "Hello, World!",
97 "This is a message box from the VMI!",
98 )),
99 )?
100 .with_pid(explorer_pid)
101 })?;
102
103 Ok(())
104}More examples
examples/windows-recipe-writefile.rs (lines 233-239)
204fn main() -> Result<(), Box<dyn std::error::Error>> {
205 let (session, _profile) = common::create_vmi_session()?;
206
207 let explorer_pid = {
208 // This block is used to drop the pause guard after the PID is found.
209 // If the `session.handle()` would be called with the VM paused, no
210 // events would be triggered.
211 let paused = session.pause_guard()?;
212
213 let vmi = paused.state();
214
215 let explorer = match common::find_process(&vmi, "explorer.exe")? {
216 Some(explorer) => explorer,
217 None => {
218 tracing::error!("explorer.exe not found");
219 return Ok(());
220 }
221 };
222
223 tracing::info!(
224 pid = %explorer.id()?,
225 object = %explorer.object()?,
226 "found explorer.exe"
227 );
228
229 explorer.id()?
230 };
231
232 session.handle(|session| {
233 UserInjectorHandler::new(
234 session,
235 recipe_factory(GuestFile::new(
236 "C:\\Users\\John\\Desktop\\test.txt",
237 "Hello, World!".as_bytes(),
238 )),
239 )?
240 .with_pid(explorer_pid)
241 })?;
242
243 Ok(())
244}examples/windows-recipe-writefile-advanced.rs (lines 336-342)
302fn main() -> Result<(), Box<dyn std::error::Error>> {
303 let (session, _profile) = common::create_vmi_session()?;
304
305 let explorer_pid = {
306 // This block is used to drop the pause guard after the PID is found.
307 // If the `session.handle()` would be called with the VM paused, no
308 // events would be triggered.
309 let paused = session.pause_guard()?;
310
311 let vmi = paused.state();
312
313 let explorer = match common::find_process(&vmi, "explorer.exe")? {
314 Some(explorer) => explorer,
315 None => {
316 tracing::error!("explorer.exe not found");
317 return Ok(());
318 }
319 };
320
321 tracing::info!(
322 pid = %explorer.id()?,
323 object = %explorer.object()?,
324 "found explorer.exe"
325 );
326
327 explorer.id()?
328 };
329
330 let mut content = Vec::new();
331 for c in 'A'..='Z' {
332 content.extend((0..2049).map(|_| c as u8).collect::<Vec<_>>());
333 }
334
335 session.handle(|session| {
336 UserInjectorHandler::new(
337 session,
338 recipe_factory(GuestFile::new(
339 "C:\\Users\\John\\Desktop\\test.txt",
340 content,
341 )),
342 )?
343 .with_pid(explorer_pid)
344 })?;
345
346 Ok(())
347}Sourcepub fn with_bridge(
vmi: &VmiSession<'_, Os>,
bridge: Bridge,
recipe: Recipe<Os, T>,
) -> Result<InjectorHandler<Os, Mode, T, Bridge>, VmiError>
pub fn with_bridge( vmi: &VmiSession<'_, Os>, bridge: Bridge, recipe: Recipe<Os, T>, ) -> Result<InjectorHandler<Os, Mode, T, Bridge>, VmiError>
Creates a new injector handler with a custom bridge for guest-host communication.
Sourcepub fn with_pid(
self,
pid: ProcessId,
) -> Result<InjectorHandler<Os, Mode, T, Bridge>, VmiError>
pub fn with_pid( self, pid: ProcessId, ) -> Result<InjectorHandler<Os, Mode, T, Bridge>, VmiError>
Restricts injection to a specific process.
Examples found in repository?
examples/windows-recipe-messagebox.rs (line 100)
64fn main() -> Result<(), Box<dyn std::error::Error>> {
65 let (session, _profile) = common::create_vmi_session()?;
66
67 let explorer_pid = {
68 // This block is used to drop the pause guard after the PID is found.
69 // If the `session.handle()` would be called with the VM paused, no
70 // events would be triggered.
71 let paused = session.pause_guard()?;
72
73 let vmi = paused.state();
74
75 let explorer = match common::find_process(&vmi, "explorer.exe")? {
76 Some(explorer) => explorer,
77 None => {
78 tracing::error!("explorer.exe not found");
79 return Ok(());
80 }
81 };
82
83 tracing::info!(
84 pid = %explorer.id()?,
85 object = %explorer.object()?,
86 "found explorer.exe"
87 );
88
89 explorer.id()?
90 };
91
92 session.handle(|session| {
93 UserInjectorHandler::new(
94 session,
95 recipe_factory(MessageBox::new(
96 "Hello, World!",
97 "This is a message box from the VMI!",
98 )),
99 )?
100 .with_pid(explorer_pid)
101 })?;
102
103 Ok(())
104}More examples
examples/windows-recipe-writefile.rs (line 240)
204fn main() -> Result<(), Box<dyn std::error::Error>> {
205 let (session, _profile) = common::create_vmi_session()?;
206
207 let explorer_pid = {
208 // This block is used to drop the pause guard after the PID is found.
209 // If the `session.handle()` would be called with the VM paused, no
210 // events would be triggered.
211 let paused = session.pause_guard()?;
212
213 let vmi = paused.state();
214
215 let explorer = match common::find_process(&vmi, "explorer.exe")? {
216 Some(explorer) => explorer,
217 None => {
218 tracing::error!("explorer.exe not found");
219 return Ok(());
220 }
221 };
222
223 tracing::info!(
224 pid = %explorer.id()?,
225 object = %explorer.object()?,
226 "found explorer.exe"
227 );
228
229 explorer.id()?
230 };
231
232 session.handle(|session| {
233 UserInjectorHandler::new(
234 session,
235 recipe_factory(GuestFile::new(
236 "C:\\Users\\John\\Desktop\\test.txt",
237 "Hello, World!".as_bytes(),
238 )),
239 )?
240 .with_pid(explorer_pid)
241 })?;
242
243 Ok(())
244}examples/windows-recipe-writefile-advanced.rs (line 343)
302fn main() -> Result<(), Box<dyn std::error::Error>> {
303 let (session, _profile) = common::create_vmi_session()?;
304
305 let explorer_pid = {
306 // This block is used to drop the pause guard after the PID is found.
307 // If the `session.handle()` would be called with the VM paused, no
308 // events would be triggered.
309 let paused = session.pause_guard()?;
310
311 let vmi = paused.state();
312
313 let explorer = match common::find_process(&vmi, "explorer.exe")? {
314 Some(explorer) => explorer,
315 None => {
316 tracing::error!("explorer.exe not found");
317 return Ok(());
318 }
319 };
320
321 tracing::info!(
322 pid = %explorer.id()?,
323 object = %explorer.object()?,
324 "found explorer.exe"
325 );
326
327 explorer.id()?
328 };
329
330 let mut content = Vec::new();
331 for c in 'A'..='Z' {
332 content.extend((0..2049).map(|_| c as u8).collect::<Vec<_>>());
333 }
334
335 session.handle(|session| {
336 UserInjectorHandler::new(
337 session,
338 recipe_factory(GuestFile::new(
339 "C:\\Users\\John\\Desktop\\test.txt",
340 content,
341 )),
342 )?
343 .with_pid(explorer_pid)
344 })?;
345
346 Ok(())
347}Trait Implementations
Sourceimpl<Os, Mode, T, Bridge> VmiHandler<Os> for InjectorHandler<Os, Mode, T, Bridge>where
Os: InjectorExecutionAdapter<Mode, T, Bridge>,
Mode: ExecutionMode,
Bridge: BridgeDispatch<Os, u64>,
impl<Os, Mode, T, Bridge> VmiHandler<Os> for InjectorHandler<Os, Mode, T, Bridge>where
Os: InjectorExecutionAdapter<Mode, T, Bridge>,
Mode: ExecutionMode,
Bridge: BridgeDispatch<Os, u64>,
Sourcetype Output = <<Os as InjectorExecutionAdapter<Mode, T, Bridge>>::Handler as VmiHandler<Os>>::Output
type Output = <<Os as InjectorExecutionAdapter<Mode, T, Bridge>>::Handler as VmiHandler<Os>>::Output
The output type of the handler.
Sourcefn handle_event(
&mut self,
vmi: VmiContext<'_, Os>,
) -> VmiEventResponse<<Os as VmiOs>::Architecture>
fn handle_event( &mut self, vmi: VmiContext<'_, Os>, ) -> VmiEventResponse<<Os as VmiOs>::Architecture>
Called for each VMI event. Read more
Sourcefn poll(
&self,
) -> Option<<InjectorHandler<Os, Mode, T, Bridge> as VmiHandler<Os>>::Output>
fn poll( &self, ) -> Option<<InjectorHandler<Os, Mode, T, Bridge> as VmiHandler<Os>>::Output>
Checks if the handler has completed. Read more
Sourcefn handle_timeout(&mut self, _session: &VmiSession<'_, Os>)
fn handle_timeout(&mut self, _session: &VmiSession<'_, Os>)
Called when the event loop times out waiting for the next event. Read more
Sourcefn handle_interrupted(&mut self, _session: &VmiSession<'_, Os>)
fn handle_interrupted(&mut self, _session: &VmiSession<'_, Os>)
Called when the event loop is interrupted by a signal. Read more
Sourcefn cleanup(&mut self, _session: &VmiSession<'_, Os>)
fn cleanup(&mut self, _session: &VmiSession<'_, Os>)
Called once before the session tears down monitoring. Read more