InjectorHandler

Struct InjectorHandler 

pub struct InjectorHandler<Os, Mode, T, Bridge = ()>
where Os: InjectorExecutionAdapter<Mode, T, Bridge>, Mode: ExecutionMode, Bridge: BridgeDispatch<Os, u64>,
{ /* private fields */ }
Available on crate features utils and injector only.
Generic injector handler that delegates to an OS- and mode-specific implementation.

Prefer the KernelInjectorHandler and UserInjectorHandler type aliases for the common case without a bridge. When a custom BridgeDispatch is needed, use this type directly with an explicit Bridge parameter.

Implementations

Source

impl<Os, Mode, T, Bridge> InjectorHandler<Os, Mode, T, Bridge>
where Os: InjectorExecutionAdapter<Mode, T, Bridge>, Mode: ExecutionMode, Bridge: BridgeDispatch<Os, u64>,

pub fn new( vmi: &VmiSession<'_, Os>, recipe: Recipe<Os, T>, ) -> Result<InjectorHandler<Os, Mode, T, Bridge>, VmiError>
where Bridge: Default,

Creates a new injector handler with a default (no-op) bridge.

Examples found in repository?
examples/windows-recipe-messagebox.rs (lines 95-101)
65fn main() -> Result<(), Box<dyn std::error::Error>> {
66    let (session, _profile) = common::create_vmi_session()?;
67
68    let explorer_pid = {
69        // This block is used to drop the pause guard after the PID is found.
70        // If the `session.handle()` would be called with the VM paused, no
71        // events would be triggered.
72        let _pause_guard = session.pause_guard()?;
73
74        let registers = session.registers(VcpuId(0))?;
75        let vmi = session.with_registers(&registers);
76
77        let explorer = match common::find_process(&vmi, "explorer.exe")? {
78            Some(explorer) => explorer,
79            None => {
80                tracing::error!("explorer.exe not found");
81                return Ok(());
82            }
83        };
84
85        tracing::info!(
86            pid = %explorer.id()?,
87            object = %explorer.object()?,
88            "found explorer.exe"
89        );
90
91        explorer.id()?
92    };
93
94    session.handle(|session| {
95        UserInjectorHandler::new(
96            session,
97            recipe_factory(MessageBox::new(
98                "Hello, World!",
99                "This is a message box from the VMI!",
100            )),
101        )?
102        .with_pid(explorer_pid)
103    })?;
104
105    Ok(())
106}
More examples
Hide additional examples
examples/windows-recipe-writefile.rs (lines 233-239)
203fn main() -> Result<(), Box<dyn std::error::Error>> {
204    let (session, _profile) = common::create_vmi_session()?;
205
206    let explorer_pid = {
207        // This block is used to drop the pause guard after the PID is found.
208        // If the `session.handle()` would be called with the VM paused, no
209        // events would be triggered.
210        let _pause_guard = session.pause_guard()?;
211
212        let registers = session.registers(VcpuId(0))?;
213        let vmi = session.with_registers(&registers);
214
215        let explorer = match common::find_process(&vmi, "explorer.exe")? {
216            Some(explorer) => explorer,
217            None => {
218                tracing::error!("explorer.exe not found");
219                return Ok(());
220            }
221        };
222
223        tracing::info!(
224            pid = %explorer.id()?,
225            object = %explorer.object()?,
226            "found explorer.exe"
227        );
228
229        explorer.id()?
230    };
231
232    session.handle(|session| {
233        UserInjectorHandler::new(
234            session,
235            recipe_factory(GuestFile::new(
236                "C:\\Users\\John\\Desktop\\test.txt",
237                "Hello, World!".as_bytes(),
238            )),
239        )?
240        .with_pid(explorer_pid)
241    })?;
242
243    Ok(())
244}
examples/windows-recipe-writefile-advanced.rs (lines 336-342)
301fn main() -> Result<(), Box<dyn std::error::Error>> {
302    let (session, _profile) = common::create_vmi_session()?;
303
304    let explorer_pid = {
305        // This block is used to drop the pause guard after the PID is found.
306        // If the `session.handle()` would be called with the VM paused, no
307        // events would be triggered.
308        let _pause_guard = session.pause_guard()?;
309
310        let registers = session.registers(VcpuId(0))?;
311        let vmi = session.with_registers(&registers);
312
313        let explorer = match common::find_process(&vmi, "explorer.exe")? {
314            Some(explorer) => explorer,
315            None => {
316                tracing::error!("explorer.exe not found");
317                return Ok(());
318            }
319        };
320
321        tracing::info!(
322            pid = %explorer.id()?,
323            object = %explorer.object()?,
324            "found explorer.exe"
325        );
326
327        explorer.id()?
328    };
329
330    let mut content = Vec::new();
331    for c in 'A'..='Z' {
332        content.extend((0..2049).map(|_| c as u8).collect::<Vec<_>>());
333    }
334
335    session.handle(|session| {
336        UserInjectorHandler::new(
337            session,
338            recipe_factory(GuestFile::new(
339                "C:\\Users\\John\\Desktop\\test.txt",
340                content,
341            )),
342        )?
343        .with_pid(explorer_pid)
344    })?;
345
346    Ok(())
347}
pub fn with_bridge( vmi: &VmiSession<'_, Os>, bridge: Bridge, recipe: Recipe<Os, T>, ) -> Result<InjectorHandler<Os, Mode, T, Bridge>, VmiError>

Creates a new injector handler with a custom bridge for guest-host communication.

pub fn with_pid( self, pid: ProcessId, ) -> Result<InjectorHandler<Os, Mode, T, Bridge>, VmiError>

Restricts injection to a specific process.

Examples found in repository?
examples/windows-recipe-messagebox.rs (line 102)
65fn main() -> Result<(), Box<dyn std::error::Error>> {
66    let (session, _profile) = common::create_vmi_session()?;
67
68    let explorer_pid = {
69        // This block is used to drop the pause guard after the PID is found.
70        // If the `session.handle()` would be called with the VM paused, no
71        // events would be triggered.
72        let _pause_guard = session.pause_guard()?;
73
74        let registers = session.registers(VcpuId(0))?;
75        let vmi = session.with_registers(&registers);
76
77        let explorer = match common::find_process(&vmi, "explorer.exe")? {
78            Some(explorer) => explorer,
79            None => {
80                tracing::error!("explorer.exe not found");
81                return Ok(());
82            }
83        };
84
85        tracing::info!(
86            pid = %explorer.id()?,
87            object = %explorer.object()?,
88            "found explorer.exe"
89        );
90
91        explorer.id()?
92    };
93
94    session.handle(|session| {
95        UserInjectorHandler::new(
96            session,
97            recipe_factory(MessageBox::new(
98                "Hello, World!",
99                "This is a message box from the VMI!",
100            )),
101        )?
102        .with_pid(explorer_pid)
103    })?;
104
105    Ok(())
106}
More examples
Hide additional examples
examples/windows-recipe-writefile.rs (line 240)
203fn main() -> Result<(), Box<dyn std::error::Error>> {
204    let (session, _profile) = common::create_vmi_session()?;
205
206    let explorer_pid = {
207        // This block is used to drop the pause guard after the PID is found.
208        // If the `session.handle()` would be called with the VM paused, no
209        // events would be triggered.
210        let _pause_guard = session.pause_guard()?;
211
212        let registers = session.registers(VcpuId(0))?;
213        let vmi = session.with_registers(&registers);
214
215        let explorer = match common::find_process(&vmi, "explorer.exe")? {
216            Some(explorer) => explorer,
217            None => {
218                tracing::error!("explorer.exe not found");
219                return Ok(());
220            }
221        };
222
223        tracing::info!(
224            pid = %explorer.id()?,
225            object = %explorer.object()?,
226            "found explorer.exe"
227        );
228
229        explorer.id()?
230    };
231
232    session.handle(|session| {
233        UserInjectorHandler::new(
234            session,
235            recipe_factory(GuestFile::new(
236                "C:\\Users\\John\\Desktop\\test.txt",
237                "Hello, World!".as_bytes(),
238            )),
239        )?
240        .with_pid(explorer_pid)
241    })?;
242
243    Ok(())
244}
examples/windows-recipe-writefile-advanced.rs (line 343)
301fn main() -> Result<(), Box<dyn std::error::Error>> {
302    let (session, _profile) = common::create_vmi_session()?;
303
304    let explorer_pid = {
305        // This block is used to drop the pause guard after the PID is found.
306        // If the `session.handle()` would be called with the VM paused, no
307        // events would be triggered.
308        let _pause_guard = session.pause_guard()?;
309
310        let registers = session.registers(VcpuId(0))?;
311        let vmi = session.with_registers(&registers);
312
313        let explorer = match common::find_process(&vmi, "explorer.exe")? {
314            Some(explorer) => explorer,
315            None => {
316                tracing::error!("explorer.exe not found");
317                return Ok(());
318            }
319        };
320
321        tracing::info!(
322            pid = %explorer.id()?,
323            object = %explorer.object()?,
324            "found explorer.exe"
325        );
326
327        explorer.id()?
328    };
329
330    let mut content = Vec::new();
331    for c in 'A'..='Z' {
332        content.extend((0..2049).map(|_| c as u8).collect::<Vec<_>>());
333    }
334
335    session.handle(|session| {
336        UserInjectorHandler::new(
337            session,
338            recipe_factory(GuestFile::new(
339                "C:\\Users\\John\\Desktop\\test.txt",
340                content,
341            )),
342        )?
343        .with_pid(explorer_pid)
344    })?;
345
346    Ok(())
347}